#include "header.h"

#include <stdio.h>
#include <stdlib.h>

#define TOOLS_C

#include "labels.h"
#include "tools.h"
#include "errors.h"

#include <string.h>

struct typelist snapshot_list[] = {
  {SNAPSHOT_SAVEFILE, "file", NULL},
#ifndef NO_PIPED_COMMANDS
  {SNAPSHOT_EXEC_CMD, "command", NULL},
  {SNAPSHOT_EXEC_CMD_ASYNC, "command_async", NULL},
#endif /* NO_PIPED_COMMANDS */
  {SNAPSHOT_SAVEFILE, NULL, NULL}}; /* default */

/* alloc_entries - allocate and initialize an entries-structure. Return NULL 
   on error. */

struct entries *alloc_entries(void)
	{
	struct entries *en;
	
	if ((en = (struct entries *)malloc(sizeof(struct entries))) == NULL)
		{
		Msg(0, "could not allocate memory for entries, probably not enough free mem");
		return NULL;
		}
	
	/* initialize to reasonable values */
	en->dimension = 0;
	en->topol = TOPOL_UNKNOWN;
	en->neigh = NEIGH_UNKNOWN;
	en->xdim = en->ydim = 0;
	en->current = en->entry = NULL;
	en->num_loaded = 0;
	en->num_entries = 0;
	en->fi = NULL;
	en->lap = 0;
	en->buffer = 0;
	en->flags.loadmode = LOADMODE_ALL;
	en->flags.totlen_known = 0;
	en->flags.random_order = 0;
	en->flags.skip_empty = 1;
	en->flags.labels_needed = (!label_not_needed(-1));
	return en;
	}

int label_not_needed(int level)
	{
	static int label_level = 0;
	
	if (level >= 0) 
		{
		label_level = level;
		}
	
	return(label_level);
	}

/* copy_entries - create a new entries structure with the same parameters
   as the original. Doesn't copy data_entrys */
struct entries *copy_entries(struct entries *entr)
	{
	struct entries *New;
	
	New = (struct entries *)alloc_entries();
	if (New == NULL)
		return NULL;
	
	New->dimension = entr->dimension;
	New->topol = entr->topol;
	New->neigh = entr->neigh;
	New->xdim = entr->xdim;
	New->ydim = entr->ydim;
	
	return New;
	}

/* rewind_entries - go to the first entry in entries list. Returns pointer
   to first data_entry. Loads data from file if it hasn't been loaded yet and
   rewinds file if we are using buffered reading. */

struct data_entry *rewind_entries(struct entries *entries, eptr *ptr)
	{
	struct data_entry *current = entries->entry;
	
	ptr->parent = entries;
	
	if (current == NULL)
		{
		/* file not loaded into memory */
		Msg(0, "rewind_entries failed\n");
		return NULL;
		}

	ptr->current = current;
	ptr->index = 0;
	entries->lap++;
	
	return current;
	}

/* next_entry - Get next entry from the entries table. Returns NULL when 
   at end of table or end of file is encountered. If loadmode is buffered,
   loads more data from file when needed. */
struct data_entry *next_entry(eptr *ptr)
	{
	struct entries *entries = ptr->parent;
	struct data_entry *next, *current;
	
	current = ptr->current;
	
	if (current == NULL)
		return NULL;
	  
	next = current->next;
	
	ptr->current = next;
	if (next) 
		ptr->index++;
	
	return next;
	}

/* close_entries - deallocates an entries-file. Frees memory allocated for
   entry-list and closes the file associated with entries if there is one. */

void close_entries(struct entries *entries)
	{
	if (entries)
		{
		/* deallocate data */
		if (entries->entry)
			free_entrys(entries->entry);
		      
		/* free memory allocated for structure */
		free(entries);
		}
	}

/* free_entrys - Free a list of entries */

void free_entrys(struct data_entry *data)
	{
	struct data_entry *tmp;
	
	for (;data != NULL; data = tmp)
		{
		tmp = data->next;
		free_entry(data);
		}
	}

/* initialize and possibly allocate room for data_entry. If entry is NULL, 
   a new entry is allocated and initialized. If entry is a pointer to an 
   old entry, the old entry is re-initialized. Return NULL on error. */

struct data_entry *init_entry(struct entries *entr, struct data_entry *entry)
	{
	clear_err();
	if (entry == NULL)
		{
		entry = (struct data_entry *)malloc(sizeof(struct data_entry));
		if (entry == NULL)
			{
			/*Msg(0, "alloc_entry");*/
			ERROR(ERR_NOMEM);
			return NULL;
			}
		
		entry->fixed = NULL;
		entry->next = NULL;
		entry->mask = NULL;
		entry->lab.label_array = NULL;
		entry->num_labs = 0;
		
		entry->points = (float*)calloc(entr->dimension, sizeof(float));
		if (entry->points == NULL)
			{
			free_entry(entry);
			Msg(0, "alloc_entry");
			ERROR(ERR_NOMEM);
			return NULL;
			}
		}
	
	/* discard mask */
	if (entry->mask)
		{
		free(entry->mask);
		entry->mask = NULL;
		}
	
	/* discard fixed point */
	if (entry->fixed)
		{
		free(entry->fixed);
		entry->fixed = NULL;
		}
	  
	clear_entry_labels(entry);
	entry->weight = 0;
	
	return entry;
	}

/* copy_entry - Copy one entry (next==NULL) */

struct data_entry *copy_entry(struct entries *entries, struct data_entry *data)
	{
	int i;
	struct data_entry *tmp;
	  
	clear_err();
	/* allocate memory for the copy */
	tmp = alloc_entry(entries);
	if (tmp == NULL)
		return NULL;
	
	/* copy data vector */
	for (i = 0; i < entries->dimension; i++)
		tmp->points[i] = data->points[i];
	
	/* copy labels */
	copy_entry_labels(tmp, data);
	
	/* copy mask */
	if (data->mask)
		{
		if ((tmp->mask = (char*)malloc(entries->dimension)) == NULL)
			{
			Msg(0, "Can't allocate memory for mask");
			free_entry(tmp);
			ERROR(ERR_NOMEM);
			return NULL;
			}
		memcpy(tmp->mask, data->mask, entries->dimension);
		}
	
	/* copy other stuff */
	tmp->weight = data->weight;
	
	/* copy fixed point */
	if (data->fixed)
		{
		if ((tmp->fixed = (struct fixpoint *)malloc(sizeof(struct fixpoint))) == NULL)
			{
			free_entry(tmp);
			ERROR(ERR_NOMEM);
			return NULL;
			}
		memcpy(tmp->fixed, data->fixed, sizeof(struct fixpoint));
		}
	tmp->next = NULL;
	
	return(tmp);
	}
	
/* free_entry - deallocates a data_entry. */

void free_entry(struct data_entry *entry)
	{
	if (entry)
		{
		if (entry->points)
			free(entry->points);
		if (entry->fixed)
			free(entry->fixed);
		if (entry->mask)
			free(entry->mask);
		clear_entry_labels(entry);
		free(entry);
		}
	}

/******************************************************************* 
 * Routines to handle files that contain the learning rate values  *
 *******************************************************************/

int alpha_read(float *alpha, long noc, char *infile)
	{
return -1;
#if SAMI
	long i;
	char basename[STR_LNG];
	FILE *fp;
	struct file_info *fi;
	   
	if (infile == NULL)
		return 0;

	strcpy(basename, infile);
	strtok(basename, ".");
	strcat(basename, ".lra");
	
	fi = open_file(basename, "r");
	if (fi == NULL) 
		{
		ifverbose(1) 
			{
			Msg(0, "Can't open alpha file %s", basename);
			}
		return(0);
		}
	
	fp = fi2fp(fi);
	
	for (i = 0; i < noc; i++) 
		{
		if (fscanf(fp, "%g\n", &(alpha[i])) < 0)
			return(0);
		}
	
	close_file(fi);
	
	return(1);
#endif
	}

int alpha_write(float *alpha, long noc, char *outfile)
	{
	long i;
	char basename[STR_LNG];
	FILE *fp;
	struct file_info *fi;
	
	if (outfile == NULL)
		return 0;
	
	strcpy(basename, outfile);
	strtok(basename, ".");
	strcat(basename, ".lra");
	
	fi = open_file(basename, "w+");
	if (fi == NULL) 
		{
		Msg(0, "Can't open alpha file %s for writing", basename);
		return 0;
		}
	
	fp = fi2fp(fi);
	
	for (i = 0; i < noc; i++) 
		{
		fprintf(fp, "%g\n", alpha[i]);
		}
	
	close_file(fi);
	return 0;
	}

void invalidate_alphafile(char *outfile)
	{
	int i;
	char basename[STR_LNG];
	FILE *fp;
	
	if (outfile == NULL)
		return;

	strcpy(basename, outfile);
	strtok(basename, ".");
	strcat(basename, ".lra");
	
	fp = fopen(basename, "r");
	if (fp != NULL) 
		{
		ifverbose(1)
		fprintf(fp, "Removing the learning rate file %s\n", basename);
		fclose(fp);
		i = remove(basename);
		if (i) 
			{
			Msg(0, "Can not remove %s", basename);
			}
		}
	}

/*---------------------------------------------------------------------------------------------*/
	
/* open_file - open a file for reading or writing. If name == NULL,
   uses standard input or output. Allowed characters in fmode are 'r'
   for reading, 'w' for writing, 'z' for compressed files and 'p' for
   piping output/input to/from a command (name is command string). 'P'
   and 'z' are not required in the fmode because they can be guessed
   from the filename: if the name and with the suffix .gz, .z or .Z
   the z-mode is assumed. If the name starts with the unix pipe
   character '|', the p-mode is assumed and the rest of the name is
   used as the command to be run. */

#define FM_READ 1
#define FM_WRITE 2

struct file_info *open_file(char *name, char *fmode)
{
#ifdef SAMI

  FILE *fp;
  struct file_info *fi = NULL;
  char buf[1000], *s, fmode2[10];
  int mode = 0;       /* FM_READ = read, FM_WRITE = write */
  int compress = 0;   /* 0 = no compression, 1 = compress */
  int piped_com = 0;  /* piped command? */
  int len;

  fi = alloc_file_info();
  if (fi == NULL)
    return NULL;


  /* copy mode string as we may change it later */
  strcpy(fmode2, fmode); 
  fmode = fmode2;

  /* get mode: read/write, compression? */

  if (strchr(fmode, 'r'))
    mode = FM_READ;
  else if (strchr(fmode, 'w'))
    mode = FM_WRITE;
  else
    {
      /* one of the above must be used */
      Msg(0, "open_file: incorrect file mode: %s for file %s", fmode, name);
      close_file(fi);
      return NULL;
    }
    
  /* compression? */
  if ((s = strchr(fmode, 'z')))
    {
      compress = 1;
      strcpy(s, s + 1); /* remove z from mode */
    }

  /* piped command? */
  if ((s = strchr(fmode, 'p')))
    {
      piped_com = 1;
      compress = 0; /* do not allow compress and pipe together */
      strcpy(s, s + 1); /* remove p from mode */
    }

  /* if filename starts with '|' use piped command */
  if (name && (name[0] == '|'))
    {
      piped_com = 1;
      compress = 0;  /* do not allow compress and pipe together */
      name++; 
    }

  if (!piped_com)
    if (check_for_compression(name))
      compress = 1;

  /* "-" as name means use stdin/out */
  if (name)
    if (strcmp("-", name) == 0)
      name = NULL;

  /* use stdin/stdout if name == NULL */
  if (name == NULL)
    {
      fp = (mode == FM_READ) ? stdin : stdout;
      /* assume that stdin/out is a regular file or that we don't need to
	 rewind it */
      name = (mode == FM_READ) ? stdin_name : stdout_name;
    }
  else 
    if (compress || piped_com)
      {
#ifndef NO_PIPED_COMMANDS
	if (compress)
	  {
	    /* compressed files */
	    sprintf(buf, (mode == FM_READ) ? 
		    uncompress_command : compress_command, 
		    name);
	    fi->flags.compressed = 1;
	  }
	else
	  {
	    /* piped commands */
	    strcpy(buf, name);
	  }
	fp = popen(buf, (mode == FM_READ) ? "r" : "w" );
	if (fp == NULL)
	  {
	    fprintf(stderr, "open_file: can't execute command '%s'\n",
		    buf);
	    perror("open_file");
	  }
	fi->flags.pipe = 1;
#else /* !NO_PIPED_COMMANDS */
	fprintf(stderr, "open_file: file %s: piped commands are not supported is this version\n", name);
	close_file(fi);
	return NULL;
#endif /* !NO_PIPED_COMMANDS */
      }
    else
      {
	fp = fopen(name, fmode);
	if (fp == NULL)
	  {
	    fprintf(stderr, 
		    "file_open: can't open file '%s' for %s, mode '%s'\n",
		    name, 
		    (mode == FM_READ) ? "reading" : "writing",
		    fmode);
	    perror("file_open");
	    close_file(fi);
	    return NULL;
	  }
      }
  fi->fp = fp;

  fi->name = NULL;

  /* copy name */
  if (name)
    {
      len = strlen(name);
      if (len > 0)
	{
	  fi->name = malloc(len + 1);
	  if (fi->name == NULL)
	    {
	      fprintf(stderr, "open_file: can't allocate mem for name\n");
	      close_file(fi);
	      return NULL;
	    }

	  strcpy(fi->name, name);
	}
    }

  return fi;

#endif
return NULL;
}

/* alloc_file_info - allocate and initialize file_info structure */

static struct file_info *alloc_file_info(void)
{
#ifdef SAMI

  struct file_info *fi;

  fi = malloc(sizeof(struct file_info));
  if (fi == NULL)
    {
      perror("alloc_file_info");
      return NULL;
    }

  /* initialize structure members to reasonable values */
  fi->name = NULL;
  fi->fp = NULL;
  fi->flags.pipe = 0;
  fi->flags.compressed = 0;
  fi->flags.eof = 0;
  fi->lineno = 0;
  return fi;

#endif
return NULL;
}

/* close_file - close file and deallocate file_info */

int close_file(struct file_info *fi)
{
#ifdef SAMI
  int retcode = 0;

  if (fi)
    {
      if (fi->name)
	free(fi->name);
      if (fi->fp)
#ifndef NO_PIPED_COMMANDS
	if (fi->flags.pipe) /* piped commands + compressed files */
	  retcode = pclose(fi->fp);
	else
#endif /* NO_PIPED_COMMANDS */
	  retcode = fclose(fi->fp);
      free(fi);
    }
  return retcode;
#endif
return 0;
}

/* check_for_compression - check if name indicates compression,
   i.e. the ending is one of .gz, .z, .Z. If suffix is found, returns
   a pointer to the dot, otherwise returns NULL */

static char *check_for_compression(char *name)
{
#ifdef SAMI
  char *s;
  
  if (name == NULL)
    return NULL;

  /* look for the last '.' in name */
  s = strrchr(name, '.');
  
  if (s == NULL) /* no suffix */
    return NULL;

  if (strcmp(s, ".gz") == 0) /* compressed with gzip */
    return s;

  if (strcmp(s, ".z") == 0)  /* compressed with gzip (older version) */
    return s;

  if (strcmp(s, ".Z") == 0)  /* compressed with compress */
    return s;
    
  /* unknown suffix */
#endif
  return NULL;
}    


/* getline - get a line from file. Returns a char * to the line (a
   static buffer), NULL on error. */
    
char *getline(struct file_info *fi)
{
#ifdef SAMI
  static char *stre = NULL;
  static long strl = 0;
  long len;
  int c;
  char *tstr;
  FILE *fp = fi2fp(fi);

  fi->error = 0;
  fi->flags.eof = 0;
  /* allocate memory for line buffer */

  if (stre == NULL) 
    {
      strl = STR_LNG;
      stre = (char *) malloc(sizeof(char) * strl);
      if (stre == NULL)
	{
	  perror("getline");
	  fi->error = ERR_NOMEM;
	  return NULL;
	}
    }

  len = 0;
  /* increment file line number */
  fi->lineno += 1;
  tstr = stre;

  while (1)
    {

      /* if line buffer is too small, increase its size */
      if (len >= strl)
	{
	  strl += STR_LNG;
	  tstr = stre = realloc(stre, sizeof(char) * strl);
	  if (stre == NULL)
	    {
	      perror("getline");
	      fi->error = ERR_NOMEM;
	      return NULL;
	    }

	  if (strl > ABS_STR_LNG) 
	    {
	      fprintf(stderr, "getline: Too long lines in file %s (max %d)\n",
		      fi->name, ABS_STR_LNG);
	      fi->error = ERR_LINETOOLONG;
	      return NULL;
	    }
	}

      /* get next character */
      c = fgetc(fp);

      /* end of line? */
      if (c == '\n')
	{
	  tstr[len] = '\0';
	  break;
	}

      /* end of file / error? */
      if (c == EOF)
	{
	  tstr[len] = '\0';

	  /* really an EOF? */
	  if (feof(fp))
	    {
	      /* we are at the end of file */
	      fi->flags.eof = 1;
	      if (len == 0)
		tstr = NULL;
	    }
	  else
	    {
	      /* read error */
	      tstr = NULL;
	      fi->error = ERR_FILEERR;
	      fprintf(stderr, "getline: read error on line %d of file %s\n", fi->lineno, fi->name);
	      perror("getline");
	    }
	  break;
	}

      tstr[len++] = c;
    }
  
  return tstr;
#endif
return NULL;
}

/*---------------------------------------------------------------------------------------------*/
/* set_teach_params - sets values in teaching parameter structure based
   on values given in codebook and data files */

int set_teach_params(struct teach_params *params, struct entries *codes, 
		     struct entries *data, long dbuffer)
	{
	int error = 0;
	  
	params->topol = codes->topol;
	params->mapdist = NULL;
	params->neigh = codes->neigh;
	params->neigh_adapt = NULL;
	
	/* these two might change when using a different variation, for example, dot product */
	
	params->winner = find_winner_euc;
	params->dist = vector_dist_euc;
	params->vector_adapt = adapt_vector;
	
	if (codes) 
		params->codes = codes;
	if (data)
		params->data = data;
		params->snapshot = NULL;
	
	return error;
	}

